<?php
/**
 * Abstract class for gateway settings.
 * Should be extended by all gateway settings classes.
 */

namespace Flo_Checkout\Gateways\Helpers;

use WC_Log_Handler_File;

/**
 * Class Abstract_Gateway_Settings
 */
abstract class Abstract_Gateway_Settings {

	/**
	 * Contains subarrays of settings for each gateway.
	 *
	 * @var array
	 */
	private static array $settings = array();

	/**
	 * Retrieve a specific value from the settings.
	 *
	 * @param string $key The key of the setting to retrieve.
	 *
	 * @return mixed|null The value of the specified setting, or null if the setting does not exist.
	 */
	public static function get( string $key ): mixed {
		if ( ! isset( self::$settings[ static::GATEWAY_ID ] ) ) {
			self::$settings[ static::GATEWAY_ID ] = get_option( 'woocommerce_' . static::GATEWAY_ID . '_settings' );
		}
		return self::$settings[ static::GATEWAY_ID ][ $key ] ?? null;
	}

	/**
	 * Update a specific setting.
	 *
	 * @param string $key The key of the setting to update.
	 * @param mixed  $value The value to update the setting with.
	 *
	 * @return void
	 */
	public static function update( string $key, mixed $value ): void {
		$settings = get_option( 'woocommerce_' . static::GATEWAY_ID . '_settings' );

		if ( ! is_array( $settings ) ) {
			$settings = array();
		}

		$settings[ $key ]                     = $value;
		self::$settings[ static::GATEWAY_ID ] = $settings;
		update_option( 'woocommerce_' . static::GATEWAY_ID . '_settings', $settings );
	}

	/**
	 * Check if the gateway is enabled.
	 *
	 * @return bool Returns true if the module is enabled, false otherwise.
	 */
	public static function is_enabled(): bool {
		return self::is_setting_enabled( 'enabled' );
	}

	/**
	 * Check if a specific setting is enabled.
	 *
	 * @param string $key The key of the setting to check.
	 *
	 * @return bool True if the setting is enabled, false otherwise.
	 */
	public static function is_setting_enabled( string $key ): bool {
		return 'yes' === self::get( $key );
	}

	/**
	 * Logs a message to the woocommerce log if logging is enabled
	 *
	 * @param string $level Log level.
	 * @param mixed  $message The message to log.
	 *
	 * @return void
	 */
	public static function log( string $level, mixed $message ): void {
		static $logger;

		if ( ! isset( $logger ) ) {
			$logger = wc_get_logger();
		}

		if ( 'yes' === self::get( 'log_enabled' ) ) {
			$logger->log(
				$level,
				wc_print_r( $message, true ),
				array( 'source' => static::GATEWAY_ID )
			);
		}
	}


	/**
	 * Retrieve the URL of the log file for the gateway.
	 *
	 * @return string The URL of the log file for the gateway.
	 */
	public static function get_log_file_url(): string {
		return add_query_arg(
			array(
				'page'     => 'wc-status',
				'tab'      => 'logs',
				'log_file' => WC_Log_Handler_File::get_log_file_name( static::GATEWAY_ID ),
			),
			admin_url( 'admin.php' )
		);
	}

	/**
	 * Retrieve the log file link if the log directory is defined.
	 *
	 * @return string The log file link if the log directory is defined, otherwise an empty string.
	 */
	public static function maybe_get_log_file_link(): string {
		if ( ! defined( 'WC_LOG_DIR' ) ) {
			return '';
		}

		return sprintf(
			'%s%s%s',
			'<a href="' . esc_url( self::get_log_file_url() ) . '" target="_blank">',
			__(
				'View Log',
				'flo-checkout-plugin'
			),
			'</a>'
		);
	}

	/**
	 * Get the settings fields for the payment method.
	 *
	 * @return array An array of form fields settings for BNPL payment method.
	 */
	abstract public static function get_settings_fields(): array;
}
