<?php

namespace Flo_Checkout\Gateways;

use Flo_Checkout\Gateways\NMI\NMI_Hash_Input;
use SimpleXMLElement;
use stdClass;
use WC_Order;
use WC_Payment_Gateway;
use WP_Error;

class NMI_Pay extends WC_Payment_Gateway {

	/**
	 * Class constructor, more about it in Step 3
	 */
	public $retry_interval;

	private $logger;
	private $logger_context;
	private $sandbox;
	private $terminal_id;
	private $transaction_key;
	private $hash_key;
	private $log_enabled;

	public function __construct() {
		$this->id                 = 'nmi_pay_gateway';
		$this->method_title       = __( 'Northmill Checkout - NMI Pay', 'flo-checkout-plugin' );
		$this->method_description = __( 'Use credit card payment with Northmill. Requires SSL in live mode.', 'flo-checkout-plugin' );
		$this->has_fields         = false;
		$this->supports           = array(
			'products',
			'subscriptions',
			'subscription_cancellation',
			'subscription_suspension',
			'subscription_reactivation',
			'subscription_amount_changes',
			'subscription_date_changes',
			'subscription_payment_method_change',
			'subscription_payment_method_change_customer',
			'subscription_payment_method_change_admin',
			'multiple_subscriptions',
			'refunds',
		);

		$this->init_form_fields();
		$this->logger         = wc_get_logger();
		$this->logger_context = array( 'source' => $this->id );

		// Get setting values
		$this->enabled         = $this->get_option( 'enabled' );
		$this->sandbox         = $this->get_option( 'sandbox' );
		$this->title           = $this->get_option( 'title' );
		$this->description     = $this->get_option( 'description' );
		$this->terminal_id     = $this->sandbox == 'no' ? $this->get_option( 'live_terminal_id' ) : $this->get_option( 'sandbox_terminal_id' );
		$this->transaction_key = $this->sandbox == 'no' ? $this->get_option( 'live_transaction_key' ) : $this->get_option( 'sandbox_transaction_key' );
		$this->hash_key        = $this->sandbox == 'no' ? $this->get_option( 'live_hash_key' ) : $this->get_option( 'sandbox_hash_key' );
		$this->log_enabled     = $this->get_option( 'log_enabled' );

		// Save settings fields
		add_action(
			'woocommerce_update_options_payment_gateways_' . $this->id,
			array(
				$this,
				'process_admin_options',
			),
			10,
			0
		);

		add_action( 'woocommerce_api_nmipaygateway', array( $this, 'nmi_callback_response' ), 10, 0 );

		add_action(
			'woocommerce_scheduled_subscription_payment_' . $this->id,
			array(
				$this,
				'scheduled_NMI_subscription_payment',
			),
			10,
			2
		);
	}

	/**
	 * Plugin options, we deal with it in Step 3 too
	 */
	public function init_form_fields() {
		$label       = __( 'Enable Logging', 'flo-checkout-plugin' );
		$description = __( 'Enable the logging of errors.', 'flo-checkout-plugin' );

		if ( defined( 'WC_LOG_DIR' ) ) {
			$nmi_pay_log_url = add_query_arg( 'tab', 'logs', add_query_arg( 'page', 'wc-status', admin_url( 'admin.php' ) ) );
			$nmi_pay_log_key = 'nmi_pay-' . sanitize_file_name( wp_hash( 'nmi_pay' ) ) . '-log';
			$nmi_pay_log_url = add_query_arg( 'log_file', $nmi_pay_log_key, $nmi_pay_log_url );

			// translators: %1$s: opening a tag, %2$s: closing a tag.
			$label .= ' | ' . sprintf( __( '%1$sView Log%2$s', 'flo-checkout-plugin' ), '<a href="' . esc_url( $nmi_pay_log_url ) . '">', '</a>' );
		}

		$this->form_fields = array(
			'enabled'                 => array(
				'title'       => __( 'Enable/Disable', 'flo-checkout-plugin' ),
				'label'       => __( 'Enable credit card payment with Northmill', 'flo-checkout-plugin' ),
				'type'        => 'checkbox',
				'description' => '',
				'default'     => 'no',
			),
			'sandbox'                 => array(
				'title'       => __( 'Sandbox', 'flo-checkout-plugin' ),
				'label'       => __( 'Enable Sandbox Mode', 'flo-checkout-plugin' ),
				'type'        => 'checkbox',
				'description' => __( 'Enabling test mode(real payments will not be taken).', 'flo-checkout-plugin' ),
				'default'     => 'yes',
			),
			'title'                   => array(
				'title'       => __( 'title', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This controls the title which the user sees during checkout within the credit card payment type.', 'flo-checkout-plugin' ),
				'default'     => __( 'Pay with credit card', 'flo-checkout-plugin' ),
			),
			'description'             => array(
				'title'       => __( 'description', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This controls the description which the user sees during checkout within the credit card payment type.', 'flo-checkout-plugin' ),
				'default'     => __( 'Pay with credit card.', 'flo-checkout-plugin' ),
			),
			'sandbox_terminal_id'     => array(
				'title'       => __( 'Sandbox - Terminal-ID', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This is the sandbox account Terminal-ID', 'flo-checkout-plugin' ),
				'default'     => '',
			),
			'sandbox_transaction_key' => array(
				'title'       => __( 'Sandbox Transaction Key', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This is the sandbox Transaction Key. Follow these instructions create accounts https://testwebmis.creditcall.com.', 'flo-checkout-plugin' ),
				'default'     => '',
			),
			'sandbox_hash_key'        => array(
				'title'       => __( 'Sandbox Hash Key', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This is hash key for sandbox transaction. This hash key can be request from eKashu Support (support@ekashu.com)', 'flo-checkout-plugin' ),
				'default'     => '',
			),
			'live_terminal_id'        => array(
				'title'       => __( 'Live - Terminal-ID', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This is the live account Terminal-ID.', 'flo-checkout-plugin' ),
				'default'     => '',
			),
			'live_transaction_key'    => array(
				'title'       => __( 'Live Transaction Key', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This is the live Transaction Key. Follow these instructions create accounts https://testwebmis.creditcall.com.', 'flo-checkout-plugin' ),
				'default'     => '',
			),
			'live_hash_key'           => array(
				'title'       => __( 'Live Hash Key', 'flo-checkout-plugin' ),
				'type'        => 'text',
				'description' => __( 'This is hash key for live transaction. This hash key can be request from eKashu Support (support@ekashu.com)', 'flo-checkout-plugin' ),
				'default'     => '',
			),
			'log_enabled'             => array(
				'title'       => __( 'Debug Log', 'flo-checkout-plugin' ),
				'label'       => $label,
				'description' => $description,
				'type'        => 'checkbox',
				'default'     => 'no',
			),
		);
	}

	/**
	 * You will need it if you want your custom credit card form, Step 4 is about it
	 */
	public function payment_fields() {
		if ( $this->description ) {
			echo wpautop( wp_kses_post( $this->description ) );
		}
	}

	/*
	 * Fields validation, more in Step 5
	*/
	public function validate_fields() {}

	/*
	* We're processing the payments here, everything about it is in Step 5
	*/
	public function process_payment( $order_id ) {

		global $woocommerce;

		$this->write_Woo_logs( 'process_payment' );

		$nminonce = wp_create_nonce( 'nmi_callback_mode' );
		update_post_meta( $order_id, '_nmi_payment_page', $nminonce );// Create security calls

		return array(
			'result'   => NMI_Hash_Input::SUCCESS,
			'redirect' => get_site_url() . '/?wc-api=NmiPayGateway&event_type=proccess&order_id=' . $order_id,
			// get_site_url() sometime it might cause an issue for network path enable site
		);
	}

	/*
	 * Callback methods
	 *
	 * @phpcs:disable WordPress.Security.NonceVerification.Missing
	 * @phpcs:disable WordPress.Security.NonceVerification.Recommended
	 */
	public function nmi_callback_response() {

		if ( $_REQUEST['event_type'] ) {
			$event_type = $_REQUEST['event_type'];

			if ( isset( $_REQUEST['order_id'] ) ) {
				$order_id = sanitize_text_field( $_REQUEST['order_id'] );
			}

			// assign order id from post object from ekashu return on fail,success or callbacks methods.
			if ( isset( $_POST[ NMI_Hash_Input::EKASHU_REFERENCE ] ) ) {
				$order_id = $_POST[ NMI_Hash_Input::EKASHU_REFERENCE ];
			}

			if ( 'success' === $_POST['ekashu_auth_result'] ) {
				$return_result = ! empty( $_POST ) && $this->hashVerification( $_POST );

				if ( $return_result ) {
					$this->successPayment( $order_id, $event_type );
				} else {
					$this->write_Woo_logs( 'hash verification failed' );
					$this->write_Woo_logs( $_REQUEST );
				}
			} elseif ( 'fail' === $event_type || 'fail_payment_nmi' === $event_type ) {
				$this->failPayment( $order_id, $event_type );
			} elseif ( 'proccess' === $event_type ) {
				$this->write_Woo_logs( 'process payment' );
				$this->processPayment( $order_id );
			}
		}
	}
	// @phpcs:enable

	public function hashVerification( $post ) {
		if ( $post['ekashu_seller_id'] != $this->terminal_id ) {
			return false;
		}

		if ( $post['ekashu_seller_key'] != substr( str_replace( ' ', '', $this->transaction_key ), 0, 8 ) ) {
			return false;
		}

		$endpoint = ( $this->sandbox == 'yes' )
			? 'https://test.ekashu.com'
			: 'https://live.ekashu.com';

		return NMI_Hash_Input::validate_hash( $post, $endpoint );
	}

	private function processPayment( $order_id ) {

		$order     = wc_get_order( $order_id );
		$metafield = get_post_meta( $order_id, '_nmi_payment_page_proccess' );

		if ( count( $metafield ) > 0 ) {
			wc_add_notice( __( 'This order already been proccess, Please try again later', 'flo-checkout-plugin' ), 'error' );
			wp_safe_redirect( wc_get_checkout_url() );
		}
		if ( $this->sandbox == 'yes' ) {
			$endpoint = 'https://test.ekashu.com';
		} else {
			$endpoint = 'https://live.ekashu.com';
		}

		$ekashu_terminal_id     = $this->terminal_id;
		$ekashu_transaction_key = substr( str_replace( ' ', '', $this->transaction_key ), 0, 8 );// Only assigning first 8 digits

		// Get Order ID and Key
		$ekashu_id = $order_id;

		// Get Order Totals $0.00
		$ekashu_currency   = $order->get_currency();
		$ekashu_total      = $order->get_total();
		$woodecimaloptions = get_option( 'woocommerce_price_num_decimals' );
		if ( $woodecimaloptions == 0 ) {
			$ekashu_total = number_format( (float) $ekashu_total, 2, '.', '' );
		} else {
			$ekashu_total = $this->formateAmount( $ekashu_total );
		}

		// Get Order User, Billing & Shipping Addresses
		$ekashu_email = $order->get_billing_email();
		$ekashu_phone = $order->get_billing_phone();

		// Get Callbacks, success and fail URLs
		$home = get_site_url();

		$ekashu_success_url          = $home . '/?wc-api=NmiPayGateway&event_type=success_payment_nmi';
		$ekashu_failure_url          = $home . '/?wc-api=NmiPayGateway&event_type=fail_payment_nmi';
		$ekashu_return_url           = $order->get_cancel_order_url();
		$ekashu_callback_success_url = $home . '/?wc-api=NmiPayGateway&event_type=success';
		$ekashu_callback_failure_url = $home . '/?wc-api=NmiPayGateway&event_type=fail';

		$site_title       = ( get_bloginfo( 'name' ) != null )
			? html_entity_decode( get_bloginfo( 'name' ) )
			: ' ';
		$site_description = ( get_bloginfo( 'description' ) != null )
			? html_entity_decode( get_bloginfo( 'description' ) )
			: ' ';

		$ekashu_pay_page_style = FLO_CHECKOUT_URL . 'assets/css/flo_pay_nmi.css';
		$ekashu_shortcut_icon  = FLO_CHECKOUT_URL . 'assets/images/northmill-fav.png';
		$ekashu_title          = 'Northmill';

		$ekashu_request_type = 'auth';
		if ( function_exists( 'wcs_order_contains_subscription' ) ) {
			$orderContainSubscription = wcs_order_contains_subscription( $order );
			if ( $orderContainSubscription === true && $ekashu_total == 0 ) {
				$ekashu_request_type = 'preauth';
			}
		}

		$hash_key = $this->hash_key;
		try {
			$hash_input = new NMI_Hash_Input(
				array(
					'ekashu_amount'                   => $ekashu_total,
					'ekashu_currency'                 => $ekashu_currency,
					'ekashu_reference'                => $ekashu_id,
					'ekashu_seller_id'                => $ekashu_terminal_id,
					'ekashu_seller_key'               => $ekashu_transaction_key,

					'ekashu_card_address_editable'    => 'true',

					'ekashu_card_address_verify'      => 'false',
					'ekashu_card_zip_code_verify'     => 'false',
					'ekashu_card_title_mandatory'     => 'false',
					'ekashu_delivery_title_mandatory' => 'false',

					'ekashu_seller_name'              => $site_title,
					'ekashu_description'              => $site_description,

					'ekashu_style_sheet'              => $ekashu_pay_page_style,
					'ekashu_shortcut_icon'            => $ekashu_shortcut_icon,
					'ekashu_title'                    => $ekashu_title,

					'ekashu_callback_include_post'    => 'true',
					'ekashu_include_post'             => 'true',
					'ekashu_auto_confirm'             => 'true',

					'ekashu_viewport'                 => 'device-width, initial-scale=1',

					'ekashu_failure_url'              => $ekashu_failure_url,
					'ekashu_success_url'              => $ekashu_success_url,
					'ekashu_callback_success_url'     => $ekashu_callback_success_url,
					'ekashu_callback_failure_url'     => $ekashu_callback_failure_url,
					'ekashu_request_type'             => $ekashu_request_type,
				)
			);
		} catch ( Exception $e ) {
			wc_add_notice( __( 'Error during checkout. Please choose another payment option.', 'flo-checkout-plugin' ), 'error' );
			wp_safe_redirect( wc_get_checkout_url() );
			exit;
		}

		$computedHash = $hash_input->get_hash( $hash_key );

		echo '<p class="payinstruction">' . __( 'Please wait! You are now redirected to a secure payment page', 'flo-checkout-plugin' ) . '</p>';
		echo '<form id="payment_page_form" action=' . $endpoint . ' method="post">';
		foreach ( $hash_input->get_form_data() as $key => $val ) {
			if ( is_null( $val ) ) {
				continue;
			}
			if ( is_bool( $val ) ) {
				echo '<input type="hidden" name="' . esc_attr( $key ) . '" value="' . ( $val ? 'true' : 'false' ) . '"/>';
			} else {
				echo '<input type="hidden" name="' . esc_attr( $key ) . '" value="' . esc_attr( $val ) . '"/>';
			}
		}
		echo '<input type="hidden" name="ekashu_hash_code" value="' . esc_attr( $computedHash ) . '"/>';
		echo '<input type="hidden" name="ekashu_return_url" value="' . $ekashu_return_url . '"/>';
		echo '<input type="hidden" name="ekashu_card_name_required" value="true"/>';
		echo '<input type="hidden" name="ekashu_card_name_mandatory" value="true"/>';
		echo '<input type="hidden" name="ekashu_card_phone_number" value="' . $ekashu_phone . '"/>';
		echo '<input type="hidden" name="ekashu_delivery_email_mandatory" value="false"/>';
		echo '<input type="hidden" name="ekashu_valid_from_required" value="false" />';
		echo '<input type="hidden" name="ekashu_card_email_address" value="' . $ekashu_email . '" />';
		echo '<input style="display:none" type="submit" value="Pay"/>';
		echo '<style>#payment_page_form::before {height: 1em; width: 1em;display: block;position: absolute;
            top: 50%;left: 50%;margin-left: -0.5em;margin-top: -0.5em;content: "";animation: spin 1s ease-in-out infinite;
            background: url("' . esc_url( FLO_CHECKOUT_URL . 'assets/images/loader.svg' ) . '") center center;background-size: cover;line-height: 1;text-align: center;
            font-size: 2em;color: rgba(0, 0, 0, 0.75)}@keyframes spin {from {transform:rotate(0deg);}to {transform:rotate(360deg);}}
            p.payinstruction{text-align:center;margin: 0 !important;}</style>';
		echo '<script type="text/javascript">
            document.getElementById("payment_page_form").submit();
            </script>
            </form>';
		update_post_meta( $order_id, '_nmi_payment_page_proccess', 'hasproccess' );// Create security calls
		wp_die();
	}

	private function failPayment( $order_id, $event_type ) {
		$order = new WC_Order( $order_id );

		$current_order_status = $order->get_status();
		if ( $current_order_status != WC_FAILED_STATUS ) {// This is stop update again the status
			$this->write_Woo_logs( 'Fail payment' );      // Logs
			$order->update_status( WC_FAILED_STATUS );
		}

		if ( $event_type == 'fail_payment_nmi' ) {
			$this->write_Woo_logs( $order_id . '- Fail payment redirect' );// Logs
			wc_add_notice( __( 'Payment failed. Please contact bank for information', 'flo-checkout-plugin' ), 'error' );
			wp_safe_redirect( wc_get_checkout_url() );
		}

		if ( $event_type == 'fail' ) {                                               // Callback fail method
			$this->write_Woo_logs( $order_id . '- Fail callback response received' );// Logs
			$returnResult[ NMI_Hash_Input::SUCCESS ] = true;
			$returnResult['errorCode']               = '200';

			wp_send_json( $returnResult, $returnResult['errorCode'] );// Send received notification to server
		}
	}

	private function successPayment( $order_id, $event_type ) {
		$order = new WC_Order( $order_id );

		update_post_meta( $order_id, NMI_Hash_Input::EKASHU_TRANSACTION_ID, $_POST[ NMI_Hash_Input::EKASHU_TRANSACTION_ID ] );// Adding transaction id

		$this->subcriptions_order( $order_id );

		if ( isset( $_POST[ NMI_Hash_Input::EKASHU_TRANSACTION_ID ] ) ) {
			$order->set_transaction_id( $_POST[ NMI_Hash_Input::EKASHU_TRANSACTION_ID ] );
		}
		$order->payment_complete();
		$order->add_order_note( __( 'Order payment received', 'flo-checkout-plugin' ) );    // New translation
		$this->write_Woo_logs( 'Payment completed' );                              // Logs
		if ( $event_type == 'success_payment_nmi' ) {
			$this->write_Woo_logs( $order_id . '- successful payment redirect' );// Logs
			WC()->cart->empty_cart();
			$thankyouURL = $order->get_checkout_order_received_url();
			wp_safe_redirect( $thankyouURL );
		}

		if ( $event_type == NMI_Hash_Input::SUCCESS ) {                                                   // Callback success method
			$this->write_Woo_logs( $order_id . '- successful callback response recived' );// Logs
			$returnResult[ NMI_Hash_Input::SUCCESS ] = true;
			$returnResult['successCode']             = '200';

			wp_send_json( $returnResult, $returnResult['successCode'] );// Send received notification to server
		}
	}

	public function subcriptions_order( $order_id ) {
		$order = new WC_Order( $order_id );

		if ( function_exists( 'wcs_order_contains_subscription' ) ) {
			$subscriptionValidate = wcs_order_contains_subscription( $order );
			if ( $subscriptionValidate === true ) {
				$this->write_Woo_logs( 'Order contain subscriptions products' );
				$subs_order_id = '';
				$subscriptions = wcs_get_subscriptions_for_order( $order_id, array( NMI_Hash_Input::ORDER_TYPE => 'any' ) );
				foreach ( $subscriptions as $subscription_obj ) {
					$subs_order_id = $subscription_obj->get_id(); // current subscribtion id
				}

				$hash_key            = $this->hash_key;
				$card_hash           = $_POST['ekashu_card_hash'];
				$card_ref            = $_POST['ekashu_card_reference'];
				$computedcardHash    = base64_encode( $subs_order_id . $card_hash );
				$computedcardrefHash = base64_encode( $hash_key . $card_ref );

				update_post_meta( $order_id, 'ekashu_enc_hash', $computedcardHash );        // Adding transaction id
				update_post_meta( $order_id, 'ekashu_enc_reference', $computedcardrefHash );// Adding transaction id

				$this->write_Woo_logs( sprintf( 'Order contain subcriptions products order Id# %s, subscription Id# %s', $order_id, $subs_order_id ) );

				unset( $computedcardHash );
				unset( $computedcardrefHash );
			}
		}
	}


	function validate_order_has_trial( $order ) {

		if ( ! $order instanceof WC_Order ) {
			return new WP_Error( 'invalid_order', 'This order is invalid.' );
		}

		// Grab all items in an order
		if ( $items = $order->get_items() ) {

			// Loop through the items
			foreach ( $items as $item_values ) {
				$product_id = $item_values->get_product_id();

				$trial_length = get_post_meta( $product_id, '_subscription_trial_length', true );
				if ( $trial_length > 0 ) {
					return true;
				}
			}

			// Otherwise, we return false
			return false;
		}
	}

	public function scheduled_NMI_subscription_payment( $amount_to_charge, $renewal_order ) {
		$reponse = $this->process_NMI_subscription_payment( $renewal_order, $amount_to_charge );// Process subcriptionS
		if ( (string) $reponse->Result->LocalResult == 0 ) {
			$renewal_order->add_order_note( __( 'Order renewal payment received', 'flo-checkout-plugin' ) );// New translation

			$paymentTrId = (string) $reponse->TransactionDetails->CardEaseReference;
			if ( isset( $paymentTrId ) ) {
				$renewal_order->payment_complete( $paymentTrId );
				update_post_meta( $renewal_order->get_id(), NMI_Hash_Input::EKASHU_TRANSACTION_ID, $paymentTrId );// Adding transaction id
			} else {
				$renewal_order->payment_complete();
			}

			$this->write_Woo_logs( 'Order renewal completed' );
		}
		if ( (string) $reponse->Result->LocalResult == 1 ) {
			$renewal_order->add_order_note(
				sprintf(
					// translators: %s: error message.
					__(
						'Order renewal is cancelled due to %s reason',
						'flo-checkout-plugin'
					),
					(string) $reponse->Result->Errors->Error
				)
			);// New translation.
			$renewal_order->update_status( 'wc-cancelled' );
			$this->write_Woo_logs(
				sprintf(
					'Order renewal is cancelled and response from xml protocol resultcode %1$s and error# %2$s.',
					$reponse->Result->LocalResult,
					$reponse->Result->Errors->Error
				)
			);
		}
		if ( (string) $reponse->Result->LocalResult == 2 ) {
			$renewal_order->add_order_note(
				sprintf(
					// translators: %s: error message.
					__( 'Order renewal pending due to %s reason', 'flo-checkout-plugin' ),
					(string) $reponse->Result->Errors->Error
				)
			);// New translation.
			$renewal_order->update_status( 'wc-pending' );
			$this->write_Woo_logs(
				sprintf(
					'Order renewal pending and response from xml protocol resultcode %s and error# %s.',
					$reponse->Result->LocalResult,
					$reponse->Result->Errors->Error
				)
			);
		}
		if ( (string) $reponse->Result->LocalResult == 9 ) {
			$renewal_order->add_order_note(
				sprintf(
						// translators: %s: error message.
					__(
						'Sorry, we are unable to process your payment at this time. Please reactivate and Please retry later. Order renewal update as pending due to %s reason',
						'flo-checkout-plugin'
					),
					(string) $reponse->Result->Errors->Error
				)
			);// New translation.
			$renewal_order->update_status( 'wc-pending' );
			$this->write_Woo_logs(
				sprintf(
					'Order renewal pending and response from xml protocol resultcode %s and error# %s.',
					(string) $reponse->Result->LocalResult,
					(string) $reponse->Result->Errors->Error
				)
			);
		}
	}

	public function retrieveParentSubcriptionsID( $order_id ) {
		$subscriptions = wcs_get_subscriptions_for_order( $order_id, array( NMI_Hash_Input::ORDER_TYPE => 'any' ) );
		foreach ( $subscriptions as $subscription_obj ) {
			$parent_subs_id = $subscription_obj->get_parent_id(); // This is subscription parent id
			if ( ! empty( $parent_subs_id ) ) {
				return $parent_subs_id;
			}
		}
	}

	public function process_NMI_subscription_payment( $order, $amount_to_charge ) {
		$hash_key       = $this->hash_key;
		$order_id       = $order->get_id();
		$order_currency = $order->get_currency();
		$subscriptions  = wcs_get_subscriptions_for_order( $order_id, array( NMI_Hash_Input::ORDER_TYPE => 'any' ) );
		$parent_id      = '';
		$subcriptionid  = '';

		$this->write_Woo_logs( sprintf( 'Start processing NMI subscription. Date time %s order-id %s', date( NMI_Hash_Input::DATETIME_FORMAT ), $order_id ) );

		foreach ( $subscriptions as $subscription_obj ) {
			$subcriptionid = $subscription_obj->get_id();        // current subscribtion id
			$parent_id     = $subscription_obj->get_parent_id(); // parent id if needed
		}

		if ( ! empty( $parent_id ) ) {
			$cardHash      = get_post_meta( $parent_id, 'ekashu_enc_hash', true );
			$cardReference = get_post_meta( $parent_id, 'ekashu_enc_reference', true );
		}

		$ekashuCardHash = str_replace( $subcriptionid, '', base64_decode( $cardHash ) );
		$ekashuCardRef  = str_replace( $hash_key, '', base64_decode( $cardReference ) );

		$this->write_Woo_logs( sprintf( 'Processing order renewal payment request for order-id %s amount %s order parent-id %s subscription id %s ', $order_id, $amount_to_charge, $parent_id, $subcriptionid ) );

		$xmlObject = $this->prepareXMLObject( $ekashuCardHash, $ekashuCardRef, $amount_to_charge, $order_currency );
		// $this->write_Woo_logs($xmlObject);
		$response = $this->StartTransaction( $xmlObject );
		libxml_use_internal_errors( true );
		$XMLtoObject = simplexml_load_string( $response );
		if ( ! $XMLtoObject ) {
			libxml_clear_errors();
			if ( $this->retry_interval <= 3 ) {
				$this->write_Woo_logs( sprintf( 'Retry NMI subscription date-time %s order-id %s', date( NMI_Hash_Input::DATETIME_FORMAT ), $order_id ) );
				sleep( $this->retry_interval );
				++$this->retry_interval;

				return $this->process_NMI_subscription_payment( $order, $amount_to_charge );
			}
			$XMLtoObject                        = new stdClass();
			$XMLtoObject->Result->LocalResult   = 9;
			$XMLtoObject->Result->Errors->Error = $response;

		}
		$this->write_Woo_logs( sprintf( 'Completed processing NMI subscription. Date-time %s order-id %s', date( NMI_Hash_Input::DATETIME_FORMAT ), $order_id ) );

		return $XMLtoObject;
	}

	public function StartTransaction( $xmlObject ) {
		if ( $this->sandbox == 'yes' ) {
			$url = 'https://test.cardeasexml.com/generic.cex';
		} else {
			$url = 'https://live.cardeasexml.com/generic.cex';
		}

		$headers = array(
			'Content-type: text/xml',
			'Content-length: ' . strlen( $xmlObject ),
			'Connection: close',
		);

		$ch = curl_init();
		curl_setopt( $ch, CURLOPT_URL, $url );
		curl_setopt( $ch, CURLOPT_RETURNTRANSFER, 1 );
		curl_setopt( $ch, CURLOPT_TIMEOUT, 10 );
		curl_setopt( $ch, CURLOPT_POST, true );
		curl_setopt( $ch, CURLOPT_POSTFIELDS, $xmlObject );
		curl_setopt( $ch, CURLOPT_HTTPHEADER, $headers );

		$data = curl_exec( $ch );
		if ( curl_errno( $ch ) ) {
			$response = curl_error( $ch );
		} else {
			curl_close( $ch );
			$response = $data;
		}

		return $response;
	}

	public function prepareXMLObject( $ekashuCardHash, $ekashuCardRef, $amount_to_charge, $order_currency ) {
		$ekashu_terminal_id     = $this->terminal_id;
		$ekashu_transaction_key = $this->transaction_key;

		$amount_to_charge  = $this->formateAmount( $amount_to_charge );
		$woodecimaloptions = get_option( 'woocommerce_price_num_decimals' );

		if ( $woodecimaloptions > 0 ) {
			$amount_to_charge = str_replace( '.', '', $amount_to_charge );
		}

		return '<?xml version="1.0" encoding="us-ascii" standalone="yes"?>
                <Request type="CardEaseXML" version="1.0.0">
                    <TransactionDetails>
                        <MessageType autoconfirm="true">Auth</MessageType>
                        <Amount>' . $amount_to_charge . '</Amount>
                        <CurrencyCode>' . $order_currency . '</CurrencyCode>
                    </TransactionDetails>
                    <TerminalDetails>
                        <TerminalID>' . $ekashu_terminal_id . '</TerminalID>
                        <TransactionKey>' . $ekashu_transaction_key . '</TransactionKey>
                        <Software version="1.0.0">CardEaseXMLClient</Software>
                    </TerminalDetails>
                    <CardDetails>
                        <Manual type="cnp">
                            <CardReference>' . $ekashuCardRef . '</CardReference>
                            <CardHash>' . $ekashuCardHash . '</CardHash>
                        </Manual>
                    </CardDetails>
                </Request>';
	}

	/**
	 * Process refund.
	 *
	 * If the gateway declares 'refunds' support, this will allow it to refund.
	 * a passed in amount.
	 *
	 * @param int    $order_id Order ID.
	 * @param float  $amount   Refund amount.
	 * @param string $reason   Refund reason.
	 *
	 * @return bool|WP_Error True or false based on success, or a WP_Error object.
	 */
	public function process_refund( $order_id, $amount_to_refund = null, $reason = '' ) {
		$url = ( 'yes' == $this->sandbox )
			? 'https://test.cardeasexml.com/generic.cex'
			: 'https://live.cardeasexml.com/generic.cex';

		if ( ! class_exists( '\SimpleXMLElement' ) ) {
			return new WP_Error(
				'flo-checkout-nmi-pay',
				sprintf(
					__(
						'%s is not installed on this server. Please contact your server administrator.',
						'flo-checkout-plugin'
					),
					'SimpleXMLElement'
				) . ' ' . __( 'Please log in to NMI to perform refunds.', 'flo-checkout-plugin' )
			);
		}
		if ( ! function_exists( 'curl_init' ) ) {
			return new WP_Error(
				'flo-checkout-nmi-pay',
				sprintf(
						// translators: %s: Curl
					__(
						'%s is not installed on this server. Please contact your server administrator.',
						'flo-checkout-plugin'
					),
					'Curl'
				) . ' ' . __( 'Please log in to NMI to perform refunds.', 'flo-checkout-plugin' )
			);
		}

		$ekashu_transaction_id = get_post_meta( $order_id, 'ekashu_transaction_id', true );
		if ( ! $ekashu_transaction_id ) {
			return new WP_Error(
				'flo-checkout-nmi-pay',
				__(
					'Ekashu transaction ID is missing.',
					'flo-checkout-plugin'
				) . ' ' . __( 'Please log in to NMI to perform refunds.', 'flo-checkout-plugin' )
			);
		}

		if ( is_null( $amount_to_refund ) ) {
			return new WP_Error(
				'flo-checkout-nmi-pay',
				__(
					'Amount to refund missing.',
					'flo-checkout-plugin'
				) . ' ' .
				__( 'Please log in to NMI to perform refunds.', 'flo-checkout-plugin' )
			);
		}

		$order          = wc_get_order( $order_id );
		$order_currency = $order->get_currency();

		$amount_to_refund  = $this->formateAmount( $amount_to_refund );
		$woodecimaloptions = get_option( 'woocommerce_price_num_decimals' );

		if ( $woodecimaloptions > 0 ) {
			$amount_to_refund = str_replace( '.', '', $amount_to_refund );
		}

		$requestXml = sprintf(
			"<?xml version='1.0' encoding='UTF-8' standalone='yes'?>
            <Request type='CardEaseXML' version='1.2.0'>
                <TransactionDetails>
                    <MessageType>Refund</MessageType>
                    <CardEaseReference>%s</CardEaseReference>
                    %s
                </TransactionDetails>
                <TerminalDetails>
                    <TerminalID>%s</TerminalID>
                    <TransactionKey>%s</TransactionKey>
                    <Software version='1.3.6'>MoreFloCheckout</Software>
                    <Component version='1.0.0'>NmiPayGateway</Component>
                </TerminalDetails>
            </Request>",
			$ekashu_transaction_id,
			"<Amount>{$amount_to_refund}</Amount><CurrencyCode>{$order_currency}</CurrencyCode>",
			$this->terminal_id,
			$this->transaction_key
		);

		$ch = curl_init();
		curl_setopt( $ch, CURLOPT_CONNECTTIMEOUT, 100 );
		curl_setopt( $ch, CURLOPT_HTTPHEADER, array( 'Content-Type: text/xml' ) );
		curl_setopt( $ch, CURLOPT_POST, true );
		curl_setopt( $ch, CURLOPT_RETURNTRANSFER, true );
		curl_setopt( $ch, CURLOPT_TIMEOUT, 60 );
		curl_setopt( $ch, CURLOPT_URL, $url );
		curl_setopt( $ch, CURLOPT_USERAGENT, 'MoreFloCheckout NmiPayGateway' );
		curl_setopt( $ch, CURLOPT_POSTFIELDS, $requestXml );

		$response = curl_exec( $ch );
		if ( curl_errno( $ch ) ) {
			$this->write_Woo_logs( 'Error refunding order ' . $order_id . '. Request: ' . var_export( $requestXml, true ) );
			$this->write_Woo_logs( 'Error refunding order ' . $order_id . '. Curl Error: ' . var_export( curl_error( $ch ), true ) );

			return new WP_Error( 'flo-checkout-nmi-pay', __( 'Unknown error while refunding.', 'flo-checkout-plugin' ) . ' ' . __( 'Please log in to NMI to perform refunds.', 'flo-checkout-plugin' ) );
		}

		curl_close( $ch );

		// Parse response
		$xml    = new SimpleXMLElement( $response );
		$result = (string) $xml->Result->LocalResult;
		if ( '0' != $result ) {
			// Errors from NMI, abort.
			$errors = array();
			foreach ( $xml->Result->Errors->Error as $error ) {
				$errors[] = sprintf(
					// translators: %1$s: error code, %2$s: error message.
					__(
						'Error %1$s: %2$s',
						'flo-checkout-plugin'
					),
					(string) $error['code'],
					(string) $error
				);
			}

			return new WP_Error( 'flo-checkout-nmi-pay', implode( '; ', $errors ) );
		}

		// Refund success, save reference.
		add_post_meta( $order_id, 'mf_checkout_nmi_refunds', $response );

		return true;
	}

	public function write_Woo_logs( $details ) {
		if ( $this->log_enabled == 'yes' ) {// record if only log enable this method to run
			$this->logger->debug( wc_print_r( '************', true ), $this->logger_context );
			$this->logger->debug( wc_print_r( $details, true ), $this->logger_context );
			$this->logger->debug( wc_print_r( '************', true ), $this->logger_context );
		}
	}

	public function formateAmount( $amount ) {
		// Adding decimal point zeros if required based on woocommerece settings
		$woodecimaloptions = get_option( 'woocommerce_price_num_decimals' );

		$nmiamountformats = array(
			0 => '00',
			1 => '0',
		);
		foreach ( $nmiamountformats as $key => $value ) {
			if ( $key == $woodecimaloptions ) {
				$amount = $amount . $value;
			}
		}
		if ( $woodecimaloptions > 0 ) {
			if ( $woodecimaloptions >= 3 ) {
				$amount = substr( $amount, 0, - ( $woodecimaloptions - 2 ) );
			}
		}

		return $amount;
	}
}
