<?php
/**
 * Swish API class.
 */

namespace Flo_Checkout\Gateways\Swish;

/**
 * Swish API class.
 */
class Swish_API {

	/**
	 * Send an API request
	 *
	 * @param array  $request_data The payer alias.
	 * @param string $path         The API endpoint path.
	 * @param string $type         (Optional) The request type (defaults to 'POST').
	 *
	 * @return array Returns an array containing the response from the API
	 */
	public static function api_request( array $request_data, string $path, string $type = 'POST' ): array {

		$end_point   = self::get_endpoint( $path );
		$data_string = wp_json_encode( $request_data, JSON_UNESCAPED_SLASHES );

		return self::send_curl_request( $end_point, $data_string, $type );
	}

	/**
	 * Get the endpoint URL for the given path
	 *
	 * @param string $path The path to append to the endpoint URL.
	 *
	 * @return string The complete URL of the endpoint.
	 */
	private static function get_endpoint( string $path ): string {
		return 'https://' . Swish_Settings::get_swish_setting( 'api_end_url' ) . '/swish-cpcapi/api/v1/' . $path;
	}

	/**
	 * Send a POST request to a specified endpoint with the given data string.
	 *
	 * @param string $end_point   The URL of the endpoint to send the request to.
	 * @param string $data_string The data string to send in the request body.
	 * @param string $type        The type of request (e.g. "POST" or "PATCH").
	 *
	 * @return array  An array with the following keys:
	 *                - status: A boolean indicating if the request was successful or not.
	 *                - response: A string with the response message.
	 *                - content: The content of the response.
	 *                - response_technical: A string with the technical information of the response.
	 *
	 * cURL offers more options than wp_remote_post, so we use cURL instead.
	 * @phpcs:disable WordPress.WP.AlternativeFunctions.curl_curl_init
	 * @phpcs:disable WordPress.WP.AlternativeFunctions.curl_curl_setopt
	 * @phpcs:disable WordPress.WP.AlternativeFunctions.curl_curl_exec
	 * @phpcs:disable WordPress.WP.AlternativeFunctions.curl_curl_error
	 * @phpcs:disable WordPress.WP.AlternativeFunctions.curl_curl_getinfo
	 * @phpcs:disable WordPress.WP.AlternativeFunctions.curl_curl_close
	 */
	private static function send_curl_request( string $end_point, string $data_string, string $type ): array {
		$extension = strtoupper( pathinfo( Swish_Settings::get_swish_setting( 'public_certificate' ), PATHINFO_EXTENSION ) );

		$cert_check_valid = self::verify_certificate(
			(string) Swish_Settings::get_swish_setting( 'root_certificate' ),
			(string) Swish_Settings::get_swish_setting( 'public_certificate' ),
			(string) Swish_Settings::get_swish_setting( 'private_certificate' ),
			$extension
		);

		if ( true !== $cert_check_valid ) {
			Swish_Settings::log( 'debug', $cert_check_valid );

			return array(
				'status'             => false,
				'response'           => __( 'Could not connect to Swish.', 'flo-checkout-plugin' ),
				'response_technical' => $cert_check_valid,
			);
		}

		$ch = curl_init();

		curl_setopt( $ch, CURLOPT_HEADER, true );
		curl_setopt( $ch, CURLOPT_RETURNTRANSFER, true );

		curl_setopt( $ch, CURLOPT_SSL_VERIFYHOST, 2 );
		curl_setopt( $ch, CURLOPT_SSL_VERIFYPEER, true );
		curl_setopt( $ch, CURLINFO_HEADER_OUT, true );

		curl_setopt( $ch, CURLOPT_CAINFO, Swish_Settings::get_swish_setting( 'root_certificate' ) );
		curl_setopt( $ch, CURLOPT_SSLCERT, Swish_Settings::get_swish_setting( 'public_certificate' ) );
		curl_setopt( $ch, CURLOPT_URL, $end_point );
		curl_setopt( $ch, CURLOPT_CUSTOMREQUEST, $type );
		curl_setopt( $ch, CURLOPT_POSTFIELDS, $data_string );

		if ( 'P12' === $extension ) {
			curl_setopt( $ch, CURLOPT_SSLCERTTYPE, 'P12' );
			curl_setopt( $ch, CURLOPT_SSLKEYPASSWD, Swish_Settings::get_swish_setting( 'private_key' ) );

			if ( ! Swish_Settings::get_swish_setting( 'private_key' ) ) {
				Swish_Settings::log( 'debug', 'Private key not set.' );
			}
		} elseif ( 'PEM' === $extension ) {
			curl_setopt( $ch, CURLOPT_SSLKEY, Swish_Settings::get_swish_setting( 'private_certificate' ) );
			if ( Swish_Settings::get_swish_setting( 'private_key' ) ) {
				curl_setopt( $ch, CURLOPT_SSLKEYPASSWD, Swish_Settings::get_swish_setting( 'private_key' ) );
			}
		}

		if ( 'POST' === $type ) {
			curl_setopt( $ch, CURLOPT_POST, true );
		}

		$content_type = 'application/json';
		if ( 'PATCH' === $type ) {
			$content_type = 'application/json-patch+json';
		}

		curl_setopt(
			$ch,
			CURLOPT_HTTPHEADER,
			array(
				'Content-Type: ' . $content_type,
				'Content-Length: ' . strlen( $data_string ),
			)
		);

		$curl_response = curl_exec( $ch );
		$curl_error    = curl_error( $ch );
		$response_code = (int) curl_getinfo( $ch, CURLINFO_HTTP_CODE );

		$header_size      = curl_getinfo( $ch, CURLINFO_HEADER_SIZE );
		$result['header'] = substr( $curl_response, 0, $header_size );
		$result['body']   = substr( $curl_response, $header_size );

		curl_close( $ch );

		if ( $curl_response ) {
			if ( $response_code >= 200 && $response_code <= 299 ) {
				preg_match( '/Location:(.*?)\n/', $result['header'], $matches );
				$location = trim( array_pop( $matches ) );
				Swish_Settings::log( 'debug', $location );
				$return_result = array(
					'status'   => true,
					'response' => $location,
					'content'  => $curl_response,
				);
			} elseif ( $response_code >= 400 && $response_code <= 499 ) {
				Swish_Settings::log( 'debug', $result );
				$errors = json_decode( $result['body'] );

				if ( is_array( $errors ) ) {
					$errors = $errors[0];
				}

				if ( isset( $errors->errorCode ) ) {
					$error_details   = ( $errors->errorCode )
						? Swish_Errors::get_message_by_code( $errors->errorCode )
						: __( 'Could not connect to Swish.', 'flo-checkout-plugin' );
					$error_technical = $error_details;
				} else {
					$error_details   = __( 'Could not connect to Swish.', 'flo-checkout-plugin' );
					$error_technical = $errors->error ?? $error_details;
				}

				Swish_Settings::log( 'debug', 'Error during call: ' . $error_technical );
				Swish_Settings::log( 'debug', 'Response code: ' . $response_code );
				$return_result = array(
					'status'             => false,
					'response'           => $error_details,
					'response_technical' => $error_technical,
				);
			} else {
				Swish_Settings::log( 'debug', 'Error during call: Unknown response code ' . $response_code );
				$return_result = array(
					'status'             => false,
					'response'           => __( 'Could not connect to Swish.', 'flo-checkout-plugin' ),
					'response_technical' => 'Unknown response code',
				);
			}
		} else {
			Swish_Settings::log( 'debug', 'Error during call: ' . $curl_error );
			$return_result = array(
				'status'             => false,
				'response'           => __( 'Could not connect to Swish.', 'flo-checkout-plugin' ),
				'response_technical' => $curl_error,
			);
			Swish_Settings::log( 'debug', 'Root cert: ' . Swish_Settings::get_swish_setting( 'root_certificate' ) );
			Swish_Settings::log( 'debug', 'Public cert: ' . Swish_Settings::get_swish_setting( 'public_certificate' ) );
			Swish_Settings::log( 'debug', 'Private cert: ' . Swish_Settings::get_swish_setting( 'private_certificate' ) );
		}

		return $return_result;
	}
	// phpcs:enable


	/**
	 * Verify a certificate
	 *
	 * @param string $root_cert    The path to the root certificate.
	 * @param string $public_cert  The path to the public certificate.
	 * @param string $private_cert The path to the private certificate.
	 * @param string $extension    The extension of the certificate (P12 or PEM).
	 *
	 * @return true|string Returns true if the certificate is valid, otherwise returns an error message
	 */
	private static function verify_certificate( string $root_cert, string $public_cert, string $private_cert, string $extension ): true|string {

		if ( ! file_exists( $root_cert ) ) {
			return 'Root certificate not found: ' . $root_cert;

		} elseif ( ! file_exists( $public_cert ) ) {
			return 'Public certificate not found: ' . $public_cert;

		} elseif ( 'P12' === $extension ) {
			// phpcs:ignore WordPress.WP.AlternativeFunctions.file_get_contents_file_get_contents
			$p12file           = file_get_contents( $public_cert ); // Public and private key in one file.
			$certs             = array();
			$cert_parse_result = openssl_pkcs12_read( $p12file, $certs, Swish_Settings::get_swish_setting( 'private_key' ) );
			if ( ! $cert_parse_result ) {
				return 'Certificate parse failed with error: ' . openssl_error_string();
			} else {
				return true;
			}
		} elseif ( 'PEM' === $extension ) {
			if ( ! file_exists( $private_cert ) ) {
				return 'Private key not found: ' . $private_cert;
			}

			return true;
			// Test gives false fails.

			$data = 'Test data';
			openssl_sign( $data, $signature, file_get_contents( $private_cert ), 'sha256WithRSAEncryption' );
			$ok = openssl_verify( $data, $signature, file_get_contents( $public_cert ) );

			if ( 1 == $ok ) {
				return true;
			} elseif ( $ok == 0 ) {
				return 'Certificates invalid';
			} else {
				return 'Certificate parse failed with error: ' . openssl_error_string();
			}
		} else {
			return 'Unknown certificate extension.' . print_r( pathinfo( Swish_Settings::get_swish_setting( 'public_certificate' ) ), true );
		}
	}
}
