<?php
/**
 * Handles Swish client functions.
 *
 * @package flo-checkout
 */

namespace Flo_Checkout\Gateways\Swish;

use WC_Order;

/**
 * Swish client class.
 */
class Swish_Client {


	/**
	 * Prepares the payment client for the given order, payer, and merchant.
	 *
	 * @param WC_Order $order The order object.
	 * @param string   $payer The payer alias.
	 *
	 * @return array The payment details.
	 */
	public static function prepare_payment_data( WC_Order $order, string $payer ): array {

		$merchant = Swish_Settings::get_swish_setting( 'payee_number' );

		$order_id = $order->get_id();

		$secret = self::get_secret( $order_id );
		if ( ! $secret ) {
			$secret = uniqid();
			self::set_secret( $order_id, $secret );
		}

		$transaction_text_array = array();

		$customer_number = $order->get_user_id();

		$items = $order->get_items();

		$product_list = array();
		foreach ( $items as $item ) {
			$product_list[] = $item->get_name();
		}

		$products = implode( ',', $product_list );

		if ( $customer_number ) {
			$transaction_text_array[] = sprintf(
				// translators: %s is the customer number.
				__( 'Customer number %s', 'flo-checkout-plugin' ),
				$customer_number
			);
		} else {
			$transaction_text_array[] = $products;
		}

		$transaction_text = mb_substr( preg_replace( '/[^a-zA-Z0-9åäöÅÄÖ :;.,?!()]+/u', '', implode( ', ', $transaction_text_array ) ), 0, 49 );

		return array(
			'payeePaymentReference' => $order->get_order_number(),
			'callbackUrl'           => self::get_callback_url(
				$order_id,
				$secret,
				$merchant,
				'payment'
			),
			'payerAlias'            => $payer,
			'payeeAlias'            => $merchant,
			'amount'                => str_replace( ',', '.', $order->get_total() ),
			'currency'              => $order->get_currency(),
			'message'               => $transaction_text,
		);
	}

	/**
	 * Prepares the payment client for a refund transaction.
	 *
	 * @param string     $transaction_id The ID of the original payment transaction.
	 * @param WC_Order   $order          The order object.
	 * @param float|null $amount         The refund amount. If null, the total order amount will be used.
	 * @param string     $reason         The reason for the refund.
	 *
	 * @return array The refund details.
	 */
	public static function prepare_refund_payment_data(
		string $transaction_id,
		WC_Order $order,
		float|null $amount,
		string $reason
	): array {
		$merchant = Swish_Settings::get_swish_setting( 'payee_number' );
		$order_id = $order->get_id();

		$secret = self::get_secret( $order_id );
		if ( ! $secret ) {
			$secret = uniqid();
			self::set_secret( $order_id, $secret );
		}

		if ( null === $amount ) {
			$amount = $order->get_total();
		}

		$transaction_text = mb_substr( preg_replace( '/[^a-zA-Z0-9åäöÅÄÖ :;.,?!()]+/u', '', $reason ), 0, 49 );

		return array(
			'payerPaymentReference'    => 'R-' . $order_id . '-' . date( 'Y-m-d H:i', current_time( 'timestamp' ) ),
			'originalPaymentReference' => $transaction_id,
			'callbackUrl'              => self::get_callback_url( $order_id, $secret, $merchant, 'refund' ),
			'payerAlias'               => $merchant,
			'amount'                   => str_replace( ',', '.', $amount ),
			'currency'                 => $order->get_currency(),
			'message'                  => $transaction_text,
		);
	}

	/**
	 * Retrieves the secret value associated with the specified order ID.
	 *
	 * @param int $order_id The ID of the order.
	 *
	 * @return string The secret value if found, or empty string if not found.
	 */
	public static function get_secret( int $order_id ): string {
		return get_post_meta( $order_id, '_swish_unique_id', true );
	}

	/**
	 * Sets the secret value associated with the specified order ID.
	 *
	 * @param int    $order_id The ID of the order.
	 * @param string $secret   The secret value to be set.
	 *
	 * @return void
	 */
	public static function set_secret( int $order_id, string $secret ): void {
		update_post_meta( $order_id, '_swish_unique_id', $secret );
	}

	/**
	 * Retrieves the callback URL for Swish payments.
	 *
	 * @param int|string $order_id The ID of the order. String 'test_connection' can be used to test the connection.
	 * @param string     $secret   The secret value associated with the order.
	 * @param string     $merchant The merchant number.
	 * @param string     $type     The event type.
	 * @param string     $wc_api   The WooCommerce API endpoint. The default value is 'swish_pay_gateway'.
	 *
	 * @return string The callback URL generated based on the given parameters.
	 */
	public static function get_callback_url(
		int|string $order_id,
		string $secret,
		string $merchant,
		string $type,
		string $wc_api = 'swish_pay_gateway'
	): string {
		$args = array(
			'wc-api'   => $wc_api,
			'event'    => $type,
			'nonce'    => self::security_code( $secret, $order_id, $merchant ),
			'order_id' => $order_id,
		);

		return add_query_arg( $args, site_url( '/' ) );
	}

	/**
	 * Verifies the callback by comparing the received nonce with the security key generated from the secret and order details.
	 *
	 * @param string      $received_nonce The nonce received from the callback.
	 * @param string|int  $order_id       The ID of the order.
	 * @param string|null $order_secret   The secret value associated with the order (optional).
	 *
	 * @return bool Returns true if the received nonce matches the security key, false otherwise.
	 */
	public static function verify_callback(
		string $received_nonce,
		string|int $order_id,
		string|null $order_secret = null
	): bool {
		$merchant = Swish_Settings::get_swish_setting( 'payee_number' );
		$secret   = $order_secret ? $order_secret : self::get_secret( $order_id );

		$security_key = self::security_code( $secret, $order_id, $merchant );

		// Remove the last character from the security key.
		$security_key = substr( $security_key, 0, - 1 );

		return $security_key === $received_nonce;
	}

	/**
	 * Generates a security code based on the provided secret, order ID, and merchant.
	 *
	 * @param string     $secret   The secret key used for generating the security code.
	 * @param string|int $order_id The ID of the order.
	 * @param string     $merchant The number of the merchant.
	 *
	 * @return string The generated security code.
	 */
	public static function security_code( string $secret, string|int $order_id, string $merchant ): string {
		// phpcs:ignore WordPress.PHP.DiscouragedPHPFunctions.obfuscation_base64_encode
		return base64_encode( sha1( $secret . $order_id . $merchant ) );
	}
}
